{% load i18n %}

<div role="tabpanel" class="tab-pane" id="contracts">
    <h3>{% translate 'Contracts' %}</h3>
    <div class="table-responsive">
        <table class="table table-striped table-width-fix" id="tab_contracts">
            <thead>
                <tr>
                    <th>{% translate 'Contract' %}</th>
                    <th>{% translate 'Type' %}</th>
                    <th>{% translate 'From' %}</th>
                    <th>{% translate 'To' %}</th>
                    <th>{% translate 'Status' %}</th>
                    <th>{% translate 'Date Issued' %}</th>
                    <th>{% translate 'Time Left' %}</th>
                    <th>{% translate 'Info' %}</th>
                    <th></th>
                </tr>
            </thead>

            <tbody></tbody>
        </table>
    </div>
    {% include 'memberaudit/partials/character_viewer/tabs/last_updated.html' with update_section=last_updates.contracts %}
</div>
